/*
 * This file is part of Baritone.
 *
 * Baritone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Baritone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Baritone.  If not, see <https://www.gnu.org/licenses/>.
 */

/*
 * This file is part of Baritone.
 *
 * Baritone is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Baritone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Baritone.  If not, see <https://www.gnu.org/licenses/>.
 */

package baritone.bot.event;

import baritone.bot.Baritone;
import baritone.bot.chunk.CachedWorldProvider;
import baritone.bot.event.events.*;
import baritone.bot.event.events.type.EventState;
import baritone.bot.event.listener.IGameEventListener;
import baritone.bot.utils.Helper;
import baritone.bot.utils.InputOverrideHandler;
import baritone.bot.utils.interfaces.Toggleable;
import net.minecraft.client.renderer.BufferBuilder;
import net.minecraft.client.renderer.GlStateManager;
import net.minecraft.client.renderer.Tessellator;
import net.minecraft.client.renderer.vertex.DefaultVertexFormats;
import net.minecraft.client.settings.KeyBinding;
import net.minecraft.world.chunk.Chunk;
import org.lwjgl.input.Keyboard;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;

/**
 * @author Brady
 * @since 7/31/2018 11:04 PM
 */
public final class GameEventHandler implements IGameEventListener, Helper {

    private final List<IGameEventListener> listeners = new ArrayList<>();

    @Override
    public final void onTick(TickEvent event) {
        dispatch(listener -> listener.onTick(event));
    }

    @Override
    public final void onPlayerUpdate(PlayerUpdateEvent event) {
        dispatch(listener -> listener.onPlayerUpdate(event));
    }

    @Override
    public final void onProcessKeyBinds() {
        InputOverrideHandler inputHandler = Baritone.INSTANCE.getInputOverrideHandler();

        // Simulate the key being held down this tick
        for (InputOverrideHandler.Input input : InputOverrideHandler.Input.values()) {
            KeyBinding keyBinding = input.getKeyBinding();

            if (inputHandler.isInputForcedDown(keyBinding) && !keyBinding.isKeyDown()) {
                int keyCode = keyBinding.getKeyCode();

                if (keyCode < Keyboard.KEYBOARD_SIZE)
                    KeyBinding.onTick(keyCode < 0 ? keyCode + 100 : keyCode);
            }
        }

        dispatch(IGameEventListener::onProcessKeyBinds);
    }

    @Override
    public final void onSendChatMessage(ChatEvent event) {
        dispatch(listener -> listener.onSendChatMessage(event));
    }

    @Override
    public final void onChunkEvent(ChunkEvent event) {
        EventState state = event.getState();
        ChunkEvent.Type type = event.getType();

        boolean isPostPopulate = state == EventState.POST
                && type == ChunkEvent.Type.POPULATE;

        // Whenever the server sends us to another dimension, chunks are unloaded
        // technically after the new world has been loaded, so we perform a check
        // to make sure the chunk being unloaded is already loaded.
        boolean isPreUnload = state == EventState.PRE
                && type == ChunkEvent.Type.UNLOAD
                && mc.world.getChunkProvider().isChunkGeneratedAt(event.getX(), event.getZ());

        if (Baritone.settings().chunkCaching.get()) {
            if (isPostPopulate || isPreUnload) {
                CachedWorldProvider.INSTANCE.ifWorldLoaded(world -> {
                    Chunk chunk = mc.world.getChunk(event.getX(), event.getZ());
                    world.queueForPacking(chunk);
                });
            }
        }

        dispatch(listener -> listener.onChunkEvent(event));
    }

    @Override
    public final void onRenderPass(RenderEvent event) {
        /*
        CachedWorldProvider.INSTANCE.ifWorldLoaded(world -> world.forEachRegion(region -> region.forEachChunk(chunk -> {
            drawChunkLine(region.getX() * 512 + chunk.getX() * 16, region.getZ() * 512 + chunk.getZ() * 16, event.getPartialTicks());
        })));
        */

        dispatch(listener -> listener.onRenderPass(event));
    }

    @Override
    public final void onWorldEvent(WorldEvent event) {
        if (Baritone.settings().chunkCaching.get()) {
            CachedWorldProvider cache = CachedWorldProvider.INSTANCE;

            switch (event.getState()) {
                case PRE:
                    cache.closeWorld();
                    break;
                case POST:
                    cache.closeWorld();
                    if (event.getWorld() != null)
                        cache.initWorld(event.getWorld());
                    break;
            }
        }

        dispatch(listener -> listener.onWorldEvent(event));
    }

    @Override
    public final void onSendPacket(PacketEvent event) {
        dispatch(listener -> listener.onSendPacket(event));
    }

    @Override
    public final void onReceivePacket(PacketEvent event) {
        dispatch(listener -> listener.onReceivePacket(event));
    }

    @Override
    public final void onQueryItemSlotForBlocks(ItemSlotEvent event) {
        dispatch(listener -> listener.onQueryItemSlotForBlocks(event));
    }

    @Override
    public void onPlayerRelativeMove(RelativeMoveEvent event) {
        dispatch(listener -> listener.onPlayerRelativeMove(event));
    }

    public final void registerEventListener(IGameEventListener listener) {
        this.listeners.add(listener);
    }

    private void dispatch(Consumer<IGameEventListener> dispatchFunction) {
        this.listeners.stream().filter(this::canDispatch).forEach(dispatchFunction);
    }

    private boolean canDispatch(IGameEventListener listener) {
        return !(listener instanceof Toggleable) || ((Toggleable) listener).isEnabled();
    }

    private void drawChunkLine(int posX, int posZ, float partialTicks) {
        GlStateManager.enableBlend();
        GlStateManager.tryBlendFuncSeparate(770, 771, 1, 0);
        GlStateManager.color(1.0F, 1.0F, 0.0F, 0.4F);
        GlStateManager.glLineWidth(2.0F);
        GlStateManager.disableTexture2D();

        Tessellator tessellator = Tessellator.getInstance();
        BufferBuilder buffer = tessellator.getBuffer();
        double d0 = mc.getRenderManager().viewerPosX;
        double d1 = mc.getRenderManager().viewerPosY;
        double d2 = mc.getRenderManager().viewerPosZ;
        buffer.begin(3, DefaultVertexFormats.POSITION);
        buffer.pos(posX - d0, 0 - d1, posZ - d2).endVertex();
        buffer.pos(posX - d0, 256 - d1, posZ - d2).endVertex();
        tessellator.draw();

        GlStateManager.enableDepth();
        GlStateManager.depthMask(true);
        GlStateManager.enableTexture2D();
        GlStateManager.disableBlend();
    }
}
